////////////////////////////////////////////////////////////////////////////////
//
// INTEL Corporation Proprietary Information
//
// This listing is supplied under the terms of a license agreement
// with INTEL Corporation and may not be copied nor disclosed except
// in accordance with the terms of that agreement.
//
// Copyright (c) 1996 Intel Corporation. All rights reserved.
//
// Module Name:		pcidev.cpp
// Abstract:		CPciDevice Implementation
// Environment:		Windows NT 3.51/4.0, MS Visual C++ 2.x
// Notes:
//
////////////////////////////////////////////////////////////////////////////////

// disable some warnings
#pragma warning( disable: 4699 )	// Note: Creating precompiled header
#pragma warning( disable: 4201 )	// nameless struct/union
#pragma warning( disable: 4514 )	// unreferenced inline function has been removed
#pragma warning( disable: 4355 )	// 'this' : used in base member initializer list
#pragma warning( disable: 4711 )	// function 'function' selected for automatic inline expansion

#include <idrivrpp.h>
#include "pcidev.h"
#include "debug.h"

////////////////////////////////////////////////////////////////////////////////
// ctor/dtor
//
CPciDevice::CPciDevice( IN CDriverObject* DriverObject )
	:	CDeviceObject( DriverObject ),
		m_Interrupt( this ),
		m_ulBusNumber( 0 ),
		m_pBaseAddress( NULL ),
		m_ulBaseLength( 0 ),
		m_ulSlotNumber( 0 )
{
	m_DeviceObject->Flags |= DO_DIRECT_IO;
	return;
}

CPciDevice::~CPciDevice( void )
{
	return;
}

////////////////////////////////////////////////////////////////////////////////
// Driver Object helper routines
//
NTSTATUS CPciDevice::OnInitialize( IN PUNICODE_STRING RegistryPath )
{
	TRACE( "CPciDevice::OnInitialize Entered." );
	
	// find the your device
	PCI_SLOT_NUMBER pciSlotNumber;
	PCI_COMMON_CONFIG pciConfig;
	if( FALSE == PciFindDevice( vendor_id, device_id, &m_ulBusNumber, &pciSlotNumber, &pciConfig ) )
	{
		TRACE( "CPciDevice::OnInitialize FAILED on call to PciFindDevice!" );
		return STATUS_NOT_FOUND;
	}

	m_ulSlotNumber = pciSlotNumber.u.AsULONG;
	
	// when running under the checked build, resources are not correclty repored
	if( STATUS_SUCCESS != DoCheckedBuildKludge( RegistryPath ) )
	{
		TRACE( "CPciDevice::OnInitialize FAILED on call to DoCheckedBuildKludge!" );
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	// Now do it the correct way
	// Get the values from the PCI_COMMON_CONFIG structure
	ULONG BusInterruptLevel = ULONG( -1 );
	GetBusInterruptData( BusInterruptLevel );

	ULONG BusInterruptVector = BusInterruptLevel;

	PHYSICAL_ADDRESS BusAddress;
	BusAddress.HighPart = 0;
	GetBusMemoryData( BusAddress.LowPart );
	m_ulBaseLength = 0x100;

	// Report the resource usage
	ULONG ListSize = sizeof( CM_RESOURCE_LIST )	+ sizeof( CM_PARTIAL_RESOURCE_DESCRIPTOR );
	PCM_RESOURCE_LIST ResourceList = PCM_RESOURCE_LIST( new( PagedPool ) CHAR[ ListSize ] );
	if( NULL == ResourceList )
	{
		TRACE( "CPciDevice::OnInitialize - FAILED to allocate Resource List!" );
		return STATUS_NO_MEMORY;
	}

	RtlZeroMemory( ResourceList, ListSize );
	ResourceList->Count = 1;
	ResourceList->List[ 0 ].InterfaceType = PCIBus;
	ResourceList->List[ 0 ].BusNumber = m_ulBusNumber;
	ResourceList->List[ 0 ].PartialResourceList.Count = 2;	// Memory and Interrupt

	PCM_PARTIAL_RESOURCE_DESCRIPTOR Descriptor = ResourceList->List[ 0 ].PartialResourceList.PartialDescriptors;
	Descriptor->Type = CmResourceTypeMemory;
	Descriptor->ShareDisposition = CmResourceShareShared;
	Descriptor->Flags = CM_RESOURCE_MEMORY_READ_WRITE;
	Descriptor->u.Memory.Start.LowPart = BusAddress.LowPart;
	Descriptor->u.Memory.Length = m_ulBaseLength;
	Descriptor++;

	Descriptor->Type = CmResourceTypeInterrupt;
	Descriptor->ShareDisposition = CmResourceShareShared;
	Descriptor->Flags = CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;
	Descriptor->u.Interrupt.Level = BusInterruptLevel;
	Descriptor->u.Interrupt.Vector = BusInterruptVector;
	Descriptor++;

	// Report the resource usage
	BOOLEAN	bConflictDetected = FALSE;
	NTSTATUS status = ReportResourceUsage( 
		m_DriverObject->GetClassName(),
		ResourceList,
		PUCHAR( Descriptor ) - PUCHAR( ResourceList ),
		FALSE,
		&bConflictDetected );

	// free up the list
	delete ResourceList;

	if( STATUS_SUCCESS != status )
	{
		TRACE( "CPciDevice::OnInitialize - FAILED on call to ReportResourceUsage!" );
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	if( bConflictDetected )
	{
		TRACE( "CPciDevice::OnInitialize - conflict detected in call to ReportResourceUsage!" );
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	// Translate the bus address and set up the interrupt and dpc
	PHYSICAL_ADDRESS PhysicalAddress;
	ULONG AddressSpace = 0;

	// get the bus-specific physical address of the wally base
	if( FALSE == HalTranslateBusAddress( PCIBus, m_ulBusNumber, BusAddress, &AddressSpace, &PhysicalAddress ) )
	{
		TRACE( "CPciDevice::OnInitialize - FAILED on call to HalTranslateBusAddress!" );
		return STATUS_NOT_FOUND;
	}

	// did AddressSpace change
	if( 0 != AddressSpace )
	{
		TRACE( "CPciDevice::OnInitialize FAILED - AddressSpace was changed by HalTranslateBusAddress!" );
		return STATUS_NO_MEMORY;
	}

	// store the base address of Wally
	if( NULL == (m_pBaseAddress = MmMapIoSpace( PhysicalAddress, m_ulBaseLength, MmNonCached )) )
	{
		TRACE( "CPciDevice::OnInitialize FAILED on call to MmMapIoSpace!" );
		return STATUS_NO_MEMORY;
	}

	// initialize the interrupt
	if( STATUS_SUCCESS != (status = m_Interrupt.Initialize( RegistryPath, PCIBus, m_ulBusNumber, BusInterruptLevel, BusInterruptVector )) )
	{
		TRACE( "CPciDevice::OnInitialize FAILED on call to CInterruptObject::Initialize!" );
		return status;
	}
	
	// initialize the DPC (returns void)
	m_Dpc.InitializeDpcRequest( this );
		
	// do the default thing
	return CDeviceObject::OnInitialize( RegistryPath );
}

NTSTATUS CPciDevice::ConnectInterrupt( void )
{
	NTSTATUS status;
	if( STATUS_SUCCESS != (status = m_Interrupt.Connect( NULL, m_Interrupt.GetIrql(), LevelSensitive, TRUE, FALSE )) )
	{
		TRACE( "CPciDevice::ConnectInterrupt FAILED - status = 0x%lu", status );
	}
	return status;
}

VOID CPciDevice::DisconnectInterrupt( void )
{
	m_Interrupt.Disconnect();
	return;
}

VOID CPciDevice::OnUnload( void )
{
	TRACE( "CPciDevice::OnUnload Entered." );

	// It is up to the derived class to disconnect the interrupt

	// remove page table entries for base address
	if( NULL != m_pBaseAddress )
	{
		MmUnmapIoSpace( m_pBaseAddress, m_ulBaseLength );
		m_pBaseAddress = NULL;
		m_ulBaseLength = 0;
	}

	// Now free up the resources
	BOOLEAN bConflictDetected = FALSE;
	CM_RESOURCE_LIST EmptyList;
	EmptyList.Count = 0;
	if( STATUS_SUCCESS !=
		ReportResourceUsage( m_DriverObject->GetClassName(), &EmptyList, sizeof ULONG, FALSE, &bConflictDetected ) )
	{
		TRACE( "CPciDevice::OnUnload FAILED on release resources in call to ReportResourceUsage!" );
	}

	// do the default thing
	CDeviceObject::OnUnload();

	return;
}

NTSTATUS CPciDevice::DoCheckedBuildKludge( IN PUNICODE_STRING RegistryPath )
{
	// has a memory address has been written to the board's PCI config space
	ULONG ConfigData;
	if( sizeof ULONG != GetBusMemoryData( ConfigData ) )
	{
		TRACE( "CPciDevice::DoCheckedBuildKludge - FAILED on call to HalGetBusDataByOffset!" );
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	// if a memory addres has been written, then return succes
	if( 0 != ConfigData )
	{
		return STATUS_SUCCESS;
	}

	// This test is for a checked build workaround where the Hal does not write
	// the address to the adapter's PCI config space.
	PCM_RESOURCE_LIST AllocatedResources;
	NTSTATUS status;
	if( STATUS_SUCCESS !=
		(status = AssignSlotResources(
			RegistryPath, 
			m_DriverObject->GetClassName(), 
			PCIBus, 
			m_ulBusNumber, 
			m_ulSlotNumber, 
			&AllocatedResources )) )
	{
		TRACE( "CPciDevice::DoCheckedBuildKludge - FAILED on call to AssignSlotResources!" );
		return status;
	}

	// get configuration manager partial resource descriptor for our board
	PCM_PARTIAL_RESOURCE_DESCRIPTOR	PartialResourceDescriptor = NULL;
	ULONG BusInterruptLevel = ULONG( -1 );
	ULONG BusInterruptVector = ULONG( -1 );
	ULONG BaseLength = 0;
	PHYSICAL_ADDRESS BusAddress;

	if( 0 < AllocatedResources->Count &&
		PCIBus == AllocatedResources->List[ 0 ].InterfaceType &&
		m_ulBusNumber == AllocatedResources->List[ 0 ].BusNumber )
	{
		// store allocated resources in local or member variables
		for( ULONG i = 0; i < AllocatedResources->List[ 0 ].PartialResourceList.Count; i++ )
		{
			switch( AllocatedResources->List[ 0 ].PartialResourceList.PartialDescriptors[ i ].Type )
			{
			case CmResourceTypeInterrupt:
				BusInterruptLevel = AllocatedResources->List[ 0 ].PartialResourceList.PartialDescriptors[ i ].u.Interrupt.Level;
				BusInterruptVector = AllocatedResources->List[ 0 ].PartialResourceList.PartialDescriptors[ i ].u.Interrupt.Vector;
				break;

			case CmResourceTypeMemory:
				BusAddress = AllocatedResources->List[ 0 ].PartialResourceList.PartialDescriptors[ i ].u.Memory.Start;
				BaseLength = AllocatedResources->List[ 0 ].PartialResourceList.PartialDescriptors[ i ].u.Memory.Length;
				break;

			default:
				break;
			}
		}
	}

	// free AllocatedResources
	ExFreePool( AllocatedResources );
	AllocatedResources = NULL;

	// test to see if we got all of the resources that we need
	if( ULONG( -1 ) == BusInterruptLevel || 0 == BaseLength )
	{
		TRACE( "DoCheckedBuildKludge FAILED - invalid AllocatedResources!" );
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}

	// Write the resources to the device for the workaround for the checked build.

	// Read the interrupt data at 0x3c
	GetBusData( ConfigData, offset_interrupt );

	// Now write the interrupt vector to the PCI config space, preserving the upper 24 bits.
	ConfigData = (ConfigData & 0xffffff00) | (BusInterruptVector & 0xff);
	SetBusData( ConfigData, offset_interrupt );

	// Write the board's memory address into the PCI config space
	SetBusData( BusAddress.LowPart, offset_memory );

	// Now free the resources since this is a kludge
	status = AssignResources( RegistryPath, m_DriverObject->GetClassName(), NULL, &AllocatedResources );
	if( STATUS_SUCCESS != status )
	{
		TRACE( "CPciDevice::DoCheckedBuildKludge - FAILED to release resoruces on call to AssignResources!" );
	}

	return status;
}
